#!/usr/bin/perl

## upic_optimum_v1.1.20101101.pl
## Linda Ballard, 01 November 2010
##
## Input:  genemapper output after hand processing
## Relevant Columns 2: Marker name (no spaces)
##                  3: Line number (alphanumeric)
##                  4: Peak size (whole number)
##   Match input file columns to this, or modify script to match file. 
##       
## Output: space-delimited file of each marker with it's combined unique peak pattern score
##         followed by 1s in each line column that has a unique peak pattern. 
##         Allows researcher to select markers based on lines to amplify.

## used in "for $marker (sort keys %linepeak_HoHoA)"
## to compare & find unique patterns
use Array::Compare; 

##function to convert line names to numbers for uniformity
sub indexArray($@)
{
 	my $s=shift;
 	$_ eq $s && return @_ while $_=pop;
 	-1;
}

my ($marker,$line,$peak) = ('','','');
my ($val,$combo_peak,$maxlines,$marker_ct) = ('','',0,0);
my (@str_compare,@combo_arr,@tmp1_arr,@tmp2_arr,@mlines) = ((),(),(),(),());
my (%linepeak_HoHoA,%pattern_HoA,%compcomb_HoA,%markerscoresum_H) = ((),(),(),());

print "UPIC\n";
print "Input Requirement:\n\tMARKER in column 2\n\tLINE in column 3\n\tPEAK in column 4\n";
#column 1 was populated by GeneMapper originally.
print "Delimited GeneMapper data filename: ";
chomp ($gmfile=<STDIN>);

open(GMIN, $gmfile) or die "Couldn't open $gmfile: $!\n";

$gmfile =~ s/\.\w+$//g;
$upic_out = $gmfile."\_upic\-optimum\.csv";
open(UPICOUT, ">$upic_out") or die "Couldn't create $upic_out\n";  
printf UPICOUT "\t\tLines with Unique Patterns\n";

while (<GMIN>){
	next if 1..1; ##skip first line (header) of file
	next if /^\s*$/; ##skip all blank lines in file
	##read all other lines into array and split into components
	chomp; @_=split;
	$marker = $_[1]; $line = $_[2]; $peak = $_[3];
			
	##populate non-redundant array of all lines input
	if (!grep /^$line$/, @mlines) { 
		push @mlines, $line;		
	}
	##make line names uniform for ensuing calculations; 
	$line_index = indexArray($line, @mlines) + 1;
	##populate hash with 3 components
	push @{$linepeak_HoHoA{$marker}{$line_index}},$peak;
}
close (GMIN);

#report to user what script found in file
$maxlines = @mlines;
print "\n\t$maxlines lines found\n";
##print lines in 2 columns to reduce length of output (fit screen)
for ($i = 0; $i <= $maxlines; $i=$i+2) {
	print "\t$mlines[$i]\t$mlines[$i+1]\n";
}

for $marker (sort keys %linepeak_HoHoA) {
	push @{$pattern_HoA{$marker}{0}},0; #initialize unused index
	$marker_ct++;
	for $line (sort keys %{ $linepeak_HoHoA{ $marker }} ) {    	
    	for $peak (0 ..$#{ $linepeak_HoHoA{$marker}{$line}} ) {
	    	$val = $linepeak_HoHoA{$marker}{$line}[$peak];
	    	##create a string to compare peak patterns with other lines within marker
	    	push @str_compare, $val;			        	
	    }
	    ##sort peak values before combining so $combo_peak 
	    ##  is same for each combination pf the same values
	    foreach my $peak (sort @str_compare) {
		    $combo_peak = $combo_peak."_".$peak;
		} 
	    $combo_arr[$line] = $combo_peak;
	    $combo_peak = ''; @str_compare = ();			   	
	}
	##insert a null value on lines with no peaks
	for $f (1 .. $maxlines) {
		if (!$combo_arr[$f]) { $combo_arr[$f] = 0;	}	
	}
	##compare string patterns
	for $d (1 .. $maxlines-1) {
		@tmp1_arr = $combo_arr[$d];
		for $e ($d+1 .. $maxlines) {
			@tmp2_arr = $combo_arr[$e];
			my $comp = Array::Compare->new;
			#3 types of string comparison results: n-null, s-same, d-different
			if($combo_arr[$d]eq '0' || $combo_arr[$e]eq '0' ) {
				push @{$compcomb_HoA{$d}},'n';
				push @{$compcomb_HoA{$e}},'n';
			}elsif($comp->compare(\@tmp1_arr, \@tmp2_arr)){
				push @{$compcomb_HoA{$d}},'s';
				push @{$compcomb_HoA{$e}},'s';
			}else {
				push @{$compcomb_HoA{$d}},'d';
				push @{$compcomb_HoA{$e}},'d';
			}					
		}	
	}
	($score,$null_ct) = (1,0);
	
	for $la (sort keys %compcomb_HoA) {
		for $f (0 ..$#{ $compcomb_HoA{$la}} ) {
			#if pattern is the same, change score to zero	
	    	if ($compcomb_HoA{$la}[$f] eq 's') {
		    	$score = 0;	
		    # if pattern is null, add to null count
		    } elsif ($compcomb_HoA{$la}[$f] eq 'n') { 
				$null_ct++;    
			}
	 	}	 	
	 	#skip lines with no peaks
	 	if($null_ct == $maxlines - 1) {
	 		$score = 0;	 
	 	}
	 	push @{$pattern_HoA{$marker}{$la}},$score; #will use this hash to get sums
	 	($score,$null_ct) = (1,0);
	}
	##sum the scores for each marker
	for my $g (0 .. $maxlines) { 
		$scoresum += $pattern_HoA{$marker}{$g}[0];
		##accumulate max possible UPIC for all markers
		$upic_sum += $pattern_HoA{$marker}{$g}[0];
	} 		
	$markerscoresum_H{$marker}=$scoresum; ##to print with results
	$scoresum = 0; @combo_arr = (); %compcomb_HoA=();
}

##print output file header row
printf UPICOUT "Marker\tScore\t";
for ($i = 0; $i <= $maxlines; $i++) {
	printf UPICOUT "$mlines[$i]\t"
}
printf UPICOUT "\n";

$x=0;
#sort so markers are output by descending order of scoresum			
foreach $marker (sort {$markerscoresum_H{$b} <=> $markerscoresum_H{$a}} (keys(%markerscoresum_H))) {
	#prints markers with scoresum > 0
	if ($markerscoresum_H{$marker}) {
		printf UPICOUT "$marker\t$markerscoresum_H{$marker}\t";
		for($i=1;$i<=$maxlines;$i++){ 
			printf UPICOUT "$pattern_HoA{$marker}{$i}[0]\t";						
		}
		printf UPICOUT "\n";
		$x++;
	}
}
#final notes for the user
printf UPICOUT "Max UPIC is $upic_sum if these $x markers chosen\n";
print "\t$marker_ct markers input\n\t$x have non-zero scores ";
print "with a combined UPIC of $upic_sum\n";	
print "\n\tResults in $upic_out\n\n";
close (UPICOUT);
exit;