#!/usr/local/bin/perl -w

## upic_v.1.0.pl
## Linda Ballard, 21 April 2009
##                USDA-ARS MSA Genomics and Bioinformatics Research Unit
##                linda.ballard at ars.usda.gov
##
##  Citation:
##		Renee S. Arias, Linda L. Ballard, Brian E. Scheffler
##		UPIC: Perl scripts to determine the number of SSR markers to run
##		Bioinformation 3(8): 352-360, 21 April 2009
##
## Input:  genemapper output after hand processing
## Required Columns 2: Marker name (no spaces)
##                  3: DNA Sample/Line Name (alphanumeric)
##                  4: Amplicon/Peak size (whole number)
## Match input file columns to this, or modify script to match file. 
##       
## Output: space-delimited file of UPIC sums greater than zero for combinations of markers
##         from 2 to user-selected maximum. 

use Math::Combinatorics;
use Array::Compare; ## used in for "$marker (sort keys %linepeak_HoHoA)"
#use Data::Dumper; ##testing
#use Benchmark::Stopwatch; ##testing

##convert line names to numbers
sub indexArray($@)
{
 	my $s=shift;
 	$_ eq $s && return @_ while $_=pop;
 	-1;
}

my @mlines=();
($c1,$c2,$c3, $top) = (1,2,3,30);
my (%compcomb_HoA, %linepeak_HoHoA, %pattern_HoA) = ((),(),());
my (@marker_arr, @combo_arr, @tmp1_arr, @tmp2_arr, @bugme) =((),(),(),(),());
my ($marker, $line, $peak) = ('','','');
my ($val, $combo_val) = ('','');
my ($maxlines, $marker_ct,$min_combos, $max_combos, $lowest_ok_UPIC) = (0,0,2,3,0);
my $min_scoresum = 1;

printf "UPIC\n";
printf "Input Requirement:\n\tMARKER in column 2\n\tSAMPLE in column 3\n\tSIZE in column 4\n";
printf "Delimited file of GeneMapper data: ";
chomp ($gmfile=<STDIN>);

##OPTIONAL: uncomment next 6 lines if input file columns are different from example
#printf "Column number of Markers: ";
#chomp ($c1=<STDIN>);  $c1--; #adjust for array
#printf "Column number of Lines: ";
#chomp ($c2=<STDIN>);  $c2--; #adjust for array
#printf "Column number of Peaks: ";
#chomp ($c3=<STDIN>);  $c3--; #adjust for array

##OPTIONAL: uncomment next lines if you want a different (30=default) number of results for each combination
#printf "How many top UPIC values to report: ";
#chomp ($top=<STDIN>);

open(GMIN, $gmfile) or die "Couldn't open $gmfile: $!\n"; 

##start timer
#my $stopwatch = Benchmark::Stopwatch->new->start;
#$stopwatch->lap('read input');

while (<GMIN>){
	next if 1..1; ##skip header
	next if /^\s*$/; ##skip blank lines
	##read GMIN into array and split into components
	chomp; @_=split;
	$marker = $_[$c1]; 	$line = $_[$c2]; 	$peak = $_[$c3];
	
	##for lines named alpha or numerically
	if (!grep /^$line/, @mlines) { 
		push @mlines, $line;		
	}
	$line_index = indexArray($line, @mlines) + 1;
	
	push @{$linepeak_HoHoA{$marker}{$line_index}},$peak;
	$maxlines = @mlines;
}
printf "\n\t$maxlines lines found\n\n";
close (GMIN);

##what's in pattern_HoA? To find out, uncomment next 2 lines & USE statement.
#	$Data::Dumper::Indent = 1; 
#	print Dumper \%linepeak_HoHoA;
($do1_ct,$minscoresum_answer) = (0,0);
#$stopwatch->lap('calculate scores');

do {
	(@marker_arr,@bugme,@combo_arr,@tmp1_arr, @tmp2_arr,%compcomb_HoA,@scoresum_arr) = ((),(),(),(),(),(),());
	$marker_ct = 0;
	if ($do1_ct > 0) {
		printf "Minimum score of lines with peaks per marker:";
		chomp ($min_scoresum = <STDIN>);
	}
	$do1_ct++;
	for $marker (sort keys %linepeak_HoHoA) {
			#printf "$marker ";
			#push @marker_arr, $marker;
			$marker_ct++;
			for $line (sort keys %{ $linepeak_HoHoA{ $marker }} ) {    	
		    	#printf "\t$line  ";
		    	for $c (0 ..$#{ $linepeak_HoHoA{$marker}{$line}} ) {
			    	#printf "$linepeak_HoHoA{$marker}{$line}[$c] ";	
			    	$val = $linepeak_HoHoA{$marker}{$line}[$c];
			    	##create a string to compare with other lines within marker
			    	push @bugme, $val;			        	
			    }
			    #sort peak values before combining so $combo_val is same for a unique combination
			    foreach my $i (sort @bugme) {
				    $combo_val = $combo_val."_".$i;
				} 
			    #printf "$combo_val\n";
			    $combo_arr[$line] = $combo_val;
			    $combo_val = ''; @bugme = ();			   	
			}
		
			##insert a null value on lines with no peaks
			for $f (1 .. $maxlines) {
				if (!$combo_arr[$f]) { 
					$combo_arr[$f] = 0;	
				}	
			}
			##compare unique combinations
			for $d (1 .. $maxlines-1) {
				@tmp1_arr = $combo_arr[$d];
				for $e ($d+1 .. $maxlines) {
					@tmp2_arr = $combo_arr[$e];
					my $comp = Array::Compare->new;
					if($combo_arr[$d]eq '0' || $combo_arr[$e]eq '0' ) {
						#printf "$d by $e n\n";
						push @{$compcomb_HoA{$d}},'n';
						push @{$compcomb_HoA{$e}},'n';
					}elsif($comp->compare(\@tmp1_arr, \@tmp2_arr)){
						#printf "$d by $e s\n";
						push @{$compcomb_HoA{$d}},'s';
						push @{$compcomb_HoA{$e}},'s';
					}else {
						#printf "$d by $e d\n";
						push @{$compcomb_HoA{$d}},'d';
						push @{$compcomb_HoA{$e}},'d';
					}					
				}	
			}
			($score,$n_ct,$scoresum) = (1,0,0);
		
			for $la (sort keys %compcomb_HoA) {
				#printf "$la ";
				for $f (0 ..$#{ $compcomb_HoA{$la}} ) {
			    	#printf "$compcomb_HoA{$la}[$f] ";	
			    	if ($compcomb_HoA{$la}[$f] eq 's') {
				    	$score = 0;	
				    } elsif ($compcomb_HoA{$la}[$f] eq 'n') { 
						$n_ct++;    
					}
			 	}
			 	if($n_ct == $maxlines-1) {
					$score = 0;	 
			 	}
			 	#printf " $score\n";
			 	push @{$pattern_HoA{$marker}{$la}},$score; #use this to get sums
			 	($score,$n_ct) = (1,0);
			}
			##only add markers to the list of combination possibilities if the sum of their score is greater than
			##  adjustible value
			for $g (0 .. $maxlines) {
				if ($pattern_HoA{$marker}{$g}[0]) { 
					$scoresum += $pattern_HoA{$marker}{$g}[0];
				}
			}	
			push @scoresum_arr, $scoresum;
			if ($scoresum >= $min_scoresum) {	#eliminate markers to reduce combinations
				push @marker_arr, $marker; 
			}
			if ($scoresum == 0) {printf "$marker\n";}
			$scoresum = 0; @combo_arr = (); %compcomb_HoA=();
	}
	$marker_arr_ct = @marker_arr;
	printf "\t$marker_arr_ct markers have scores \>\= $min_scoresum\n\n";

	##generate counts of scoresums on the first iteration of do-until loop
	if ($do1_ct == 1) {
		my ($c_ct,$scoresum_ct) = (0,0);
		printf "\n\tCt\tCumulativeCt\tScore\n";
		for ($msc = $maxlines; $msc >= 0; $msc--) {
			for (@scoresum_arr) {
				$scoresum_ct++ if $_ == $msc;
			}
			$c_ct += $scoresum_ct;
			printf "\t$scoresum_ct\t\t$c_ct\t$msc\n";
			$scoresum_ct = 0;
		}
		printf "\n";
	}
	printf "Would you like to adjust the minimum scoresum? (0 = yes   1 = no) ";
	chomp ($minscoresum_answer = <STDIN>);
}until ($minscoresum_answer == 1);
	
my ($maxcombo_answer,$do2_ct, $total_mpc) = (0,0,0);
do {
	##computation warning
	## $marker_arr_ct = n
	## $max_combos = k && $p = k
	$do2_ct++;
	printf "\nInformative Combinations:\nMin value of k (2 - $marker_arr_ct)? ";
	chomp ($min_combos=<STDIN>);
	printf "Max value of k (not greater than $marker_arr_ct)? ";
	chomp ($max_combos=<STDIN>);
	
	##input error check
	while ($min_combos > $marker_arr_ct || $min_combos < 2) {
		printf "Please enter a min value of k from 2 to $marker_arr_ct: ";
		chomp ($min_combos=<STDIN>);
	}
	while ($max_combos > $marker_arr_ct || $max_combos < 2) {
		printf "Please enter a max value of k from 2 to $marker_arr_ct: ";
		chomp ($max_combos=<STDIN>);
	}
	for ($p = $min_combos; $p <= $max_combos; $p++) {
		$max_possible_combos = factorial($marker_arr_ct)/(factorial($p)* factorial($marker_arr_ct - $p));
		if($do2_ct == 1) {printf "\t$max_possible_combos combinations of $p\n";}
		$total_mpc += $max_possible_combos;
	}
	printf "\n\t$total_mpc possible combos of $min_combos to $max_combos\n";
	($max_possible_combos,$total_mpc) = (0,0);
	printf "Would you like to adjust min or max k? (0 = yes   1 = no): ";
    chomp ($maxcombo_answer = <STDIN>);	
} until ($maxcombo_answer == 1);

my ($sum,$flag,$upic_sum, $upic_ct) = (0,0,0,0);
@tmp1_arr = ();

$lowest_ok_UPIC = $max_combos * $min_scoresum;
printf "Lowest UPIC allowed (suggest $lowest_ok_UPIC)? ";
chomp ($lowest_ok_UPIC=<STDIN>);

$gmfile =~ s/\.\w+$//g;
$upic_out = $gmfile."\_upic".$min_combos."-".$max_combos.".csv";
open(UPICOUT, ">$upic_out") or die "Couldn't create $upic_out\n"; 
printf UPICOUT "UPIC\tM1";
for $x ($min_combos .. $max_combos) {
	printf UPICOUT "\tM".$x;	
}
printf UPICOUT "\n";

##Begin UPIC calculations
#$stopwatch->lap('calculate UPIC');

for ($i=$min_combos; $i<=$max_combos; $i++) {
	printf UPICOUT "$i\_Marker\_Combinations\n";
	$ct = 0;
  	my $combinat = Math::Combinatorics->new(count => $i,data => [@marker_arr],);  
  	printf "------------\n";
  	while(my @combo = $combinat->next_combination){ 
    	##loop thru combo array and use its markers to grab scores by line and sum
    	## if even 1 sum is zero, the final UPIC value for that combo is zero
    	## get sums from push @{$pattern_HoA{$marker}{$la}},$score;
    	for $k (1 .. $maxlines ) { #loop thru each line score in %pattern_HoA   
    		foreach my $mark (@combo) {
	 			$sum += $pattern_HoA{$mark}{$k}[0];	
	 			if ($k == 1) {
	 				push @tmp1_arr, $mark;
 				}
			}
			#skip this combination to save time
			if ($sum == 0) {
				$flag = 1;
			}
			$upic_sum += $sum;	
			$sum = 0;	
		}
				
		if($flag) {
			$upic_sum = 0;
		}
		if($upic_sum >= $lowest_ok_UPIC) { #reducing volume of UPIC sums to save memory
			$upic_ct++;
			#printf "\n$upic_sum\n";
			#toss sorted results into a hash so we can return top $top values in descending order
			#join marker combos to form key, results will be value
			$combokey = join('-',@tmp1_arr);
			push @{$combocombo_HoA{$combokey}},$upic_sum;
		}
		@tmp1_arr =();
    	$ct++; $flag = 0; $upic_sum = 0; $combokey = '';
  	}
  	$top_ct=0;
  		
	#sort the valid upics by score, return the $top top scoring combinations
  	foreach $value (sort {$combocombo_HoA{$b}[0] cmp $combocombo_HoA{$a}[0] } keys %combocombo_HoA) {
	  	if ($top_ct < $top) {
    		#print "$value $combocombo_HoA{$value}[0]\n";
    		$upic_sum = $combocombo_HoA{$value}[0];
    		@tmp1_arr = split (/-/, $value);
    		printf UPICOUT "$upic_sum \t@tmp1_arr\n";
    		$top_ct++;
		}
		$upic_sum = 0; @tmp1_arr=();
	}
  	printf UPICOUT "\n";
	print "\n$ct possible combinations of $i from $marker_arr_ct markers. $upic_ct UPIC scores GTE $lowest_ok_UPIC\n\n";
	
	$upic_ct = 0; %combocombo_HoA = ();
}
#printf UPICOUT "\n\n";
#printf UPICOUT $stopwatch->stop->summary;
#printf $stopwatch->stop->summary;

#report
print "\n\nTop $top results per combo set printed to: $upic_out\n";

close (UPICOUT);
exit;